#include <zephyr/kernel.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/i2c.h>
#include <math.h>

#include "triccaboard.h"
#include "triccagantry.h"
#include "wellplateprotocols.h"

void gantry_setup() {
    // Enable motors
    motor_switch(NUM_MOTOR_Z0, 0);
    motor_switch(NUM_MOTOR_Z1, 0);
    motor_switch(NUM_MOTOR_Z2, 0);
    motor_switch(NUM_MOTOR_Z3, 0);
    motor_switch(NUM_MOTOR_XY0, 0);
    motor_switch(NUM_MOTOR_XY1, 0);
}

void stepZ(double numSteps, int dir, double delay_us) {
    enum Mux_type type = MOTOR;
    for(double i=0.0; i<numSteps; i++) {
        // Select motor pair 1 and step
        mux_select(NUM_MOTOR_Z0, type);
        gpio_pin_set_dt(&dirs, dir);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        mux_select(NUM_MOTOR_Z2, type);
        // TODO: Step back motors double the amount of the front for *some reason
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);

        // Select motor pair 2 and step
        mux_select(NUM_MOTOR_Z1, type);
        gpio_pin_set_dt(&dirs, !dir);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        mux_select(NUM_MOTOR_Z3, type);
        // TODO: Step back motors double the amount of the front for *some reason
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);

        k_usleep(delay_us);
    }
}

void stepZ0(double numSteps, int dir, double delay_us) {
    enum Mux_type type = MOTOR;
    for(double i=0.0; i<numSteps; i++) {
        // Select motor pair 1 and step
        mux_select(NUM_MOTOR_Z0, type);
        gpio_pin_set_dt(&dirs, dir);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        k_usleep(delay_us);
    }
}

void stepZ1(double numSteps, int dir, double delay_us) {
    enum Mux_type type = MOTOR;
    for(double i=0.0; i<numSteps; i++) {
        mux_select(NUM_MOTOR_Z1, type);
        gpio_pin_set_dt(&dirs, !dir);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        k_usleep(delay_us);
    }
}

void stepZ2(double numSteps, int dir, double delay_us) {
    enum Mux_type type = MOTOR;
    for(double i=0.0; i<numSteps; i++) {
        mux_select(NUM_MOTOR_Z2, type);
        gpio_pin_set_dt(&dirs, dir);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        k_usleep(delay_us);
    }
}

void stepZ3(double numSteps, int dir, double delay_us) {
    enum Mux_type type = MOTOR;
    for(double i=0.0; i<numSteps; i++) {
        mux_select(NUM_MOTOR_Z3, type);
        gpio_pin_set_dt(&dirs, !dir);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        gpio_pin_toggle_dt(&steps);
        k_usleep(delay_us);
    }
}

void stepX(double steps, int dir, double delay_us) {
    for (double i=0.0; i<steps; i++) {
        motor_step(NUM_MOTOR_XY0, 1, dir, 0);
        motor_step(NUM_MOTOR_XY1, 1, dir, delay_us);
    }
}

void stepY(double steps, int dir, double delay_us) {
    for (double i=0.0; i<steps; i++) {
        motor_step(NUM_MOTOR_XY0, 1, dir, 0);
        motor_step(NUM_MOTOR_XY1, 1, !dir, delay_us);
    }
}

void step(enum gantry_axis axis, double steps, int dir, double delay_us) {
    if (axis == X_AXIS_MACHINE) {
        stepX(steps, dir, delay_us);
    } else if (axis == Y_AXIS_MACHINE) {
        stepY(steps, dir, delay_us);
    } else if (axis == Z_AXIS_MACHINE) {
        stepZ(steps, dir, delay_us);
    }
}

void gantry_move(struct gantry_move_unit move) {
    double usteps_x = 0.0;
    double usteps_y = 0.0;
    double usteps_z = 0.0;
    double dist_x = 0.0;
    double dist_y = 0.0;
    double dist_z = 0.0;
    switch(move.axis) {
        case X_AXIS_MACHINE:
            dist_x = move.dist_um;
            break;
        case Y_AXIS_MACHINE:
            dist_y = move.dist_um;
            break;
        case Z_AXIS_MACHINE:
            dist_z = move.dist_um;
            break;
        // TODO: Re-implement "corrected axis'"
        default:
            break;
    };
    // #usteps_xy = move_dist um / 40 mm/rot * 1 mm / 1000 um * 200 steps/rot * 8 usteps/step
    // #usteps_z = move_dist um / 40 mm/rot * 1 mm / 1000 um * 200 steps/rot * 8 usteps * 80 / 16 (gear ratio)
    // NOTE: Implicit round down due to for loop
    // TODO: track inaccuracies and take correcting steps when possible
    usteps_x = dist_x / 25.0;
    usteps_y = dist_y / 25.0;
    usteps_z = dist_z / 10.0;

    while (1) {
        if(usteps_x > 0.0) {
            stepX(1, move.dir, XY_STEP_DELAY_US);
            usteps_x -= 1.0;
        }
        if (usteps_y > 0.0)  {
            stepY(1, move.dir, XY_STEP_DELAY_US);
            usteps_y -= 1.0;
        }
        if (usteps_z > 0.0) {
            stepZ(1, move.dir, Z_STEP_DELAY_US);
            usteps_z -= 1.0;
        }
        if ((usteps_x <= 0.0) && (usteps_y <= 0.0) && (usteps_z <= 0.0)) {
            break;
        }
    }
}

void gantry_move_exe(struct gantry_move_unit* moves, int num_moves) {
    for (int i=0; i<num_moves; i++) {
        gantry_move(moves[i]);
    }
}

void gantry_home() {
    gantry_home_z();
    gantry_home_x();
    gantry_home_y();
}

double gantry_home_x() {
    double steps = 0.0;
    while(button_status(X_LIMIT)) {
        stepX(1, 1, XY_STEP_DELAY_US);
        steps += 1.0;
    }
    gantry_curr_coor.x_pos = 350.0;
    return steps;
}

double gantry_home_y() {
    double steps = 0.0;
    while(button_status(Y_LIMIT)) {
        stepY(1, 1, XY_STEP_DELAY_US);
        steps += 1.0;
    }
    gantry_curr_coor.y_pos = 350.0;
    return steps;
}

double gantry_home_z() {
    double steps = 0.0;
    while (button_status(Z0_LIMIT) && button_status(Z1_LIMIT) && button_status(Z2_LIMIT) && button_status(Z3_LIMIT)) {
        stepZ(1, 1, Z_STEP_DELAY_US);
    }
    while (button_status(Z0_LIMIT) || button_status(Z1_LIMIT) || button_status(Z2_LIMIT) || button_status(Z3_LIMIT)) {
        if (button_status(Z0_LIMIT)) {
            stepZ0(1, 1, Z_STEP_DELAY_US);
        }
        if (button_status(Z1_LIMIT)) {
            stepZ1(1, 1, Z_STEP_DELAY_US);
        }
        if (button_status(Z2_LIMIT)) {
            stepZ2(1, 1, Z_STEP_DELAY_US);
        }
        if (button_status(Z3_LIMIT)) {
            stepZ3(1, 1, Z_STEP_DELAY_US);
        }
        steps += 1.0;
    }
    gantry_curr_coor.z_pos = 135.0;
    return steps;
}

void gantry_home_return() {
    double steps_z = gantry_home_z();
    double steps_x = gantry_home_x();
    double steps_y = gantry_home_y();
    gantry_coor_init();    

    struct gantry_coor prevpos = gantry_curr_coor;
    prevpos.x_pos -= 0.025 * steps_x;
    prevpos.y_pos -= 0.025 * steps_y;
    prevpos.z_pos -= 0.01 * steps_z;

    gantry_goto_coor(prevpos);
}

void gantry_goto_coor(struct gantry_coor coor) {
    struct gantry_move_unit move[3];
    move[0].dir = 1;
    move[1].dir = 1;
    move[2].dir = 1;

    // move z-axis first to prevent bed collision
    move[0].axis = Z_AXIS_MACHINE;
    move[1].axis = X_AXIS_MACHINE;
    move[2].axis = Y_AXIS_MACHINE;

    move[0].dist_um = (coor.z_pos - gantry_curr_coor.z_pos);
    move[1].dist_um = (coor.x_pos - gantry_curr_coor.x_pos);
    move[2].dist_um = (coor.y_pos - gantry_curr_coor.y_pos);

    for (int i=0; i<3; i++) {
        // Convert mm to um
        move[i].dist_um  = move[i].dist_um * 1000.0;
        // Negative distance means we switch directions
        if (move[i].dist_um < 0.0) {
            move[i].dir = 0;
            move[i].dist_um = -move[i].dist_um;
        }
    }
    gantry_move_exe(move, 3);
    gantry_curr_coor.x_pos = coor.x_pos;
    gantry_curr_coor.y_pos = coor.y_pos;
    gantry_curr_coor.z_pos = coor.z_pos;
}

void gantry_goto_coor_d(double x, double y, double z) {
    struct gantry_coor coor;
    coor.x_pos = x;
    coor.y_pos = y;
    coor.z_pos = z;
    gantry_goto_coor(coor);
}
void gantry_set_delay_xy(double step_delay_us) {
    XY_STEP_DELAY_US = step_delay_us;
}

void gantry_set_delay_z(double step_delay_us) {
    Z_STEP_DELAY_US = step_delay_us;
}
